import multiprocessing
import functools
import numpy as np
# from aston.trace.MathSeries import savitzkygolay
from aston.trace.math_traces import movingaverage


def simple_peak_find(s, init_slope=500, start_slope=500, end_slope=200,
                     min_peak_height=50, max_peak_width=1.5):
    """
    Given a Series, return a list of tuples indicating when
    peaks start and stop and what their baseline is.
    [(t_start, t_end, hints) ...]
    """
    point_gap = 10

    def slid_win(itr, size=2):
        """Returns a sliding window of size 'size' along itr."""
        itr, buf = iter(itr), []
        for _ in range(size):
            try:
                buf += [next(itr)]
            except StopIteration:
                return
        for new_item in itr:
            yield buf
            buf = buf[1:] + [new_item]
        yield buf

    # TODO: check these smoothing defaults
    y, t = s.values, s.index.astype(float)
    smooth_y = movingaverage(y, 9)
    dxdt = np.gradient(smooth_y) / np.gradient(t)
    # dxdt = -savitzkygolay(ts, 5, 3, deriv=1).y / np.gradient(t)

    init_slopes = np.arange(len(dxdt))[dxdt > init_slope]
    if len(init_slopes) == 0:
        return []
    # get the first points of any "runs" as a peak start
    # runs can have a gap of up to 10 points in them
    peak_sts = [init_slopes[0]]
    peak_sts += [j for i, j in slid_win(init_slopes, 2) if j - i > 10]
    peak_sts.sort()

    en_slopes = np.arange(len(dxdt))[dxdt < -end_slope]
    if len(en_slopes) == 0:
        return []
    # filter out any lone points farther than 10 away from their neighbors
    en_slopes = [en_slopes[0]]
    en_slopes += [i[1] for i in slid_win(en_slopes, 3)
                  if i[1] - i[0] < point_gap or i[2] - i[1] < point_gap]
    en_slopes += [en_slopes[-1]]
    # get the last points of any "runs" as a peak end
    peak_ens = [j for i, j in slid_win(en_slopes[::-1], 2)
                if i - j > point_gap] + [en_slopes[-1]]
    peak_ens.sort()
    # avals = np.arange(len(t))[np.abs(t - 0.675) < 0.25]
    # print([i for i in en_slopes if i in avals])
    # print([(t[i], i) for i in peak_ens if i in avals])

    peak_list = []
    pk2 = 0
    for pk in peak_sts:
        # don't allow overlapping peaks
        if pk < pk2:
            continue

        # track backwards to find the true start
        while dxdt[pk] > start_slope and pk > 0:
            pk -= 1

        # now find where the peak ends
        dist_to_end = np.array(peak_ens) - pk
        pos_end = pk + dist_to_end[dist_to_end > 0]
        for pk2 in pos_end:
            if (y[pk2] - y[pk]) / (t[pk2] - t[pk]) > start_slope:
                # if the baseline beneath the peak is too large, let's
                # keep going to the next dip
                peak_list.append({'t0': t[pk], 't1': t[pk2]})
                pk = pk2
            elif t[pk2] - t[pk] > max_peak_width:
                # make sure that peak is short enough
                pk2 = pk + np.abs(t[pk:] - t[pk] - max_peak_width).argmin()
                break
            else:
                break
        else:
            # if no end point is found, the end point
            # is the end of the timeseries
            pk2 = len(t) - 1

        if pk == pk2:
            continue
        pk_hgt = max(y[pk:pk2]) - min(y[pk:pk2])
        if pk_hgt < min_peak_height:
            continue
        peak_list.append({'t0': t[pk], 't1': t[pk2]})
    return peak_list


def wavelet_peak_find(s, min_snr=1., assume_sig=4., min_length=8.0,
                      max_dist=4.0, gap_thresh=2.0):
    # this import is here to let scipy 0.9.0 at least
    # load this module
    import scipy.signal._peak_finding as spf

    y, t = s.values, s.index

    widths = np.linspace(1, 100, 200)
    cwtm = spf.cwt(y, spf.ricker, widths)
    ridges = spf._identify_ridge_lines(cwtm, widths / max_dist, gap_thresh)
    filt_ridges = spf._filter_ridge_lines(cwtm, ridges, min_length=cwtm.shape[0] / min_length, min_snr=min_snr)  # noqa

    # the next code is just to visualize how this works
    # import matplotlib.pyplot as plt
    # ctr_x, ctr_y = np.meshgrid(t, widths)
    # ctr_y *= (t[1] - t[0])
    # plt.contourf(ctr_x, ctr_y, cwtm)
    # #plt.imshow(cwtm) #, extent=(widths[0], widths[-1], times[0], times[-1]))
    # for l in ridges:
    #     plt.plot(t[l[1]], l[0] * 0.5 * (t[1] - t[0]), 'k-')
    # for l in filt_ridges:
    #     plt.plot(t[l[1]], l[0] * 0.5 * (t[1] - t[0]), 'r-')
    # #plt.plot(peaks_t, peaks_w, 'k*')  # not working
    # plt.show()

    # loop through the ridges and find the point of maximum
    # intensity on the ridge and save its characteristics
    peak_list = []
    for i, l in enumerate(filt_ridges):
        pl = np.argmax([cwtm[j, k] for j, k in zip(l[0], l[1])])
        peak_w = widths[l[0][pl]] * 0.5 * (t[1] - t[0])
        peak_amp = cwtm[l[0][pl], l[1][pl]] / (widths[l[0]][pl] ** 0.5)
        peak_t = t[l[1][pl]]
        t0, t1 = peak_t - assume_sig * peak_w, peak_t + assume_sig * peak_w
        peak_list.append({'t0': t0, 't1': t1, 'x': peak_t,
                          'h': peak_amp, 'w': peak_w})
    return peak_list


def stat_slope_peak_find(ts):
    t = ts.index

    dx = np.gradient(ts.values)
    dx2 = np.gradient(dx)

    adx = np.average(dx)
    adx2 = np.average(dx2)
    l_i = -2

    # old loop checked for concavity too; prob. not necessary
    # for i in np.arange(len(t))[dx>adx+np.std(dx[abs(dx2)<adx2+np.std(dx2)])]:

    peak_list = []
    # loop through all of the points that have a slope
    # outside of one std. dev. from average
    for i in np.arange(len(t))[dx > adx + np.std(dx)]:
        if i - l_i == 1:
            l_i = i
            continue

        # track backwards to find where this peak started
        pt0 = None
        for j in range(i - 1, 0, -1):
            if dx[j] < adx or dx2[j] < adx2:
                pt0 = t[j]
                break

        # track forwards to find where it ends
        pt1 = None
        neg = 0
        for j in range(i, len(t)):
            if dx[j] < adx:
                neg += 1
            if neg > 3 and dx[j] > adx:  # and x[j]<ax:
                pt1 = t[j]
                break

        if pt0 is not None and pt1 is not None:
            peak_list += [{'t0': pt0, 't1': pt1}]

    return peak_list


def event_peak_find(s, events, adjust_times=False):
    if adjust_times:
        # for the following, we need to assume ts is constantly spaced
        y, t = s.values, s.index

        # convert list of events into impulses that will correlate
        # with spikes in the derivative (such as peak beginning & ends)
        pulse_y = np.zeros(len(t))
        for hints in events:
            st_t, en_t = hints['t0'], hints['t1']
            pulse_y[np.argmin(np.abs(t - st_t))] = 1.
            pulse_y[np.argmin(np.abs(t - en_t))] = -1.
        cor = np.correlate(pulse_y, np.gradient(y), mode='same')

        # apply weighting to cor to make "far" correlations less likely
        cor[:len(t) // 2] *= np.logspace(0., 1., len(t) // 2)
        cor[len(t) // 2:] *= np.logspace(1., 0., len(t) - len(t) // 2)

        shift = (len(t) // 2 - cor.argmax()) * (t[1] - t[0])
        new_evts = []
        for hint in events:
            new_hint = hints.copy()
            new_hint['t0'] += shift
            new_hint['t1'] += shift
            new_evts.append(new_hint)
        return new_evts
    else:
        return events


def _peak_find_mpwrap(ts, peak_find, fopts):
    tpks = peak_find(ts, **fopts)
    for pk in tpks:
        pk['pf'] = peak_find.__name__
    return tpks


def find_peaks(tss, pf_f, f_opts={}, mp=False):
    f = functools.partial(_peak_find_mpwrap, peak_find=pf_f, fopts=f_opts)
    if mp:
        po = multiprocessing.Pool()
        peaks_found = po.map(f, tss)
        po.close()
        po.join()
    else:
        peaks_found = list(map(f, tss))
    return peaks_found


def find_peaks_as_first(tss, pf_f, f_opts={}, mp=False):
    peaks_found = [_peak_find_mpwrap(tss[0], pf_f, f_opts)]
    for ts in tss[1:]:
        tpks = []
        for p in peaks_found[0]:
            # FIXME: needs to be updated to work with pandas

            # move peak times over to correspond to first integrated
            # signal: similar to, but different from Ricci et al '94
            old_pk_ts = tss[0].twin((p['t0'], p['t1']))
            old_t = old_pk_ts.index[old_pk_ts.values.argmax()]
            new_pk_ts = ts.twin((p['t0'], p['t1']))
            off = new_pk_ts.index[new_pk_ts.values.argmax()] - old_t
            new_p = p.copy()
            new_p['t0'] = p['t0'] + off
            new_p['t1'] = p['t1'] + off
            tpks.append(new_p)
        peaks_found.append(tpks)
    return peaks_found



import re
import os.path as op
from collections import defaultdict
import numpy as np
from scipy.optimize import leastsq
from aston.peak import Peak
from aston.peak.peak_models import gaussian
from aston.trace import Chromatogram


def read_amdis_list(db, filename):

    def get_val(line, cols, key):
        return line.split('\t')[cols.index(key)]

    cmp_lvl = 2  # number of directory levels to compare
    # TODO: does this work for agilent files only?
    mapping = defaultdict(list)
    with open(filename, 'r') as f:
        cols = f.readline().split('\t')
        for line in f:
            filename = get_val(line, cols, 'FileName')
            fn = op.splitext('/'.join(filename.split('\\')[-cmp_lvl:]))[0]
            # find if filtered filename overlaps with anything in the db
            for dt in db.children_of_type('file'):
                if fn in '/'.join(dt.rawdata.split(op.sep)):
                    break
            else:
                continue
            info = {}
            info['name'] = get_val(line, cols, 'Name')
            info['p-s-time'] = get_val(line, cols, 'RT')
            info['p-s-area'] = get_val(line, cols, 'Area')
            ts = Chromatogram(np.array([np.nan]), np.array([np.nan]), [''])
            mapping[dt] += [Peak(info, ts)]
    with db:
        for dt in mapping:
            dt.children += mapping[dt]


def read_peaks(db, filename, ftype='isodat'):
    if ftype is None:
        with open(filename, 'r') as f:
            header = f.readline()
            if 'd 13C/12C[per mil]vs. VPDB' in header:
                ftype = 'isodat'
            else:
                ftype = 'amdis'
    if ftype == 'amdis':
        delim = '\t'
        cvtr = {'name': 'name',
                'p-s-time': 'rt',
                'p-s-area': 'area'}
    elif ftype == 'isodat':
        delim = ','
        cvtr = {'name': 'peak nr.',
                'p-s-time': 'rt[s]',
                'p-s-area': 'area all[vs]',
                'p-s-width': 'width[s]',
                'p-s-d13c': 'd 13c/12c[per mil]vs. vpdb',
                'p-s-d18o': 'd 18o/16o[per mil]vs. vsmow'}
    headers = None
    mapping = defaultdict(list)
    ref_pk_info = {}

    def get_val(line, cols, key):
        return line.split(delim)[cols.index(key)]

    with open(filename, 'r') as f:
        for line in f:
            if bool(re.match('filename' + delim,
                             line, re.I)) or headers is None:
                headers = line.lower().split(',')
                continue
            fn = get_val(line, headers, 'filename')
            if ftype == 'amdis':
                # AMDIS has '.FIN' sufffixes and other stuff, so
                # munge Filename to get it into right format
                cmp_lvl = 2
                fn = op.splitext('/'.join(fn.split('\\')[-cmp_lvl:]))[0]
            # find if filtered filename overlaps with anything in the db
            for dt in db.children_of_type('file'):
                if fn in '/'.join(dt.rawdata.split(op.sep)):
                    break
            else:
                continue
            info = {}
            # load all the predefined fields
            for k in cvtr:
                info[k] = get_val(line, headers, cvtr[k])

            # create peak shapes for plotting
            if ftype == 'isodat':
                rt = float(info['p-s-time']) / 60.
                width = float(info['p-s-width']) / 60.
                t = np.linspace(rt - width, rt + width)
                data = []
                for ion in ['44', '45', '46']:
                    area = float(get_val(line, headers,
                                         'rarea ' + ion + '[mvs]')) / 60.
                    # bgd = float(get_val(line, headers, \
                    #                       'bgd ' + ion + '[mv]'))
                    height = float(get_val(line, headers,
                                           'ampl. ' + ion + '[mv]'))
                    # save the height at 44 into the info for linearity
                    if ion == '44':
                        info['p-s-ampl44'] = height
                    # 0.8 is a empirical number to make things look better
                    data.append(gaussian(t, x=rt, w=0.5 * area / height,
                                         h=height))
                # save info if this is the main ref gas peak
                if info['name'].endswith('*'):
                    ref_pk_info[dt] = info
                ts = Chromatogram(np.array(data).T, t, [44, 45, 46])
            else:
                ts = Chromatogram(np.array([np.nan]), np.array([np.nan]), [''])
            mapping[dt] += [Peak(info, ts)]
    # do drift correction
    if ftype == 'isodat':
        for dt in mapping:
            ref_pks = []
            hgt44 = ref_pk_info[dt]['p-s-ampl44']
            d18o = float(ref_pk_info[dt]['p-s-d18o'])
            d13c = float(ref_pk_info[dt]['p-s-d13c'])
            for pk in mapping[dt]:
                # if the d18o and height are similar, it's a ref peak
                if abs(pk.info['p-s-ampl44'] - hgt44) < 10. and \
                   abs(float(pk.info['p-s-d18o']) - d18o) < 2.:
                    ref_pks.append(pk)

            # get out the dd13C values and times for the ref gas peaks
            d13cs = [float(pk.info['p-s-d13c']) for pk in ref_pks]
            dd13cs = np.array(d13cs) - d13c
            rts = [float(pk.info['p-s-time']) for pk in ref_pks]

            # try to fit a linear model through all of them
            p0 = [d13cs[0], 0]

            def errfunc(p, x, y):
                return p[0] + p[1] * x - y

            try:
                p, succ = leastsq(errfunc, p0, args=(np.array(rts), dd13cs))
            except Exception:
                p = p0
            # apply the linear model to get the dd13C linearity correction
            # for a given time and add it to the value of this peak
            for pk in mapping[dt]:
                pk.info['p-s-d13c'] = str(-errfunc(p,
                                                   float(pk.info['p-s-time']),
                                                   float(pk.info['p-s-d13c'])))

    # save everything
    with db:
        for dt in mapping:
            dt.children += mapping[dt]



"""
Functions which mathematically manipulate TimeSeries.
"""
import struct
import zlib
import numpy as np
import scipy.ndimage
from aston.trace import Chromatogram, Trace


def series_from_str(val, times, name=''):
    # TODO: generate this without needing the times? just the time length
    # we can store time-series data as a list of timepoints
    # in certain info fields and query it here
    def is_num(x):
        # stupid function to determine if something is a number
        try:
            float(x)
            return True
        except ValueError:
            return False

    if ',' in val:
        # turn the time list into a dictionary
        tpts = dict([tpt.split(':') for tpt in val.split(',')])
        # get the valid times out
        valid_x = [v for v in tpts if is_num(v)]
        # generate arrays from them
        x = np.array([float(v) for v in valid_x])
        y = np.array([float(tpts[v]) for v in valid_x])
        srt_ind = np.argsort(x)
        if 'S' in tpts:
            # there's a "S"tart value defined
            d = np.interp(times, x[srt_ind], y[srt_ind], float(tpts['S']))
        else:
            d = np.interp(times, x[srt_ind], y[srt_ind])
    elif is_num(val):
        d = np.ones(times.shape) * float(val)
    else:
        d = np.ones(times.shape) * np.nan
    return Trace(d, times, name=name)


def fft(ts):
    """
    Perform a fast-fourier transform on a Trace
    """
    t_step = ts.index[1] - ts.index[0]
    oc = np.abs(np.fft.fftshift(np.fft.fft(ts.values))) / len(ts.values)
    t = np.fft.fftshift(np.fft.fftfreq(len(oc), d=t_step))
    return Trace(oc, t)


def ifft(ic, t):
    raise NotImplementedError
#    ic = np.fft.ifft(np.fft.fftshift(ic * len(ic)))# / len(ic)


def noisefilter(arr, bandwidth=0.2):
    # adapted from http://glowingpython.blogspot.com/
    # 2011/08/fourier-transforms-and-image-filtering.html
    i = np.fft.fftshift(np.fft.fft(arr))  # entering to frequency domain
    # fftshift moves zero-frequency component to the center of the array
    p = np.zeros(len(i), dtype=complex)
    c1 = len(i) / 2  # spectrum center
    r = float(bandwidth)  # percent of signal to save
    r = int((r * len(i)) / 2)  # convert to coverage of the array
    for i in range(c1 - r, c1 + r):
        p[i] = i[i]  # frequency cutting
    return np.real(np.fft.ifft(np.fft.ifftshift(p)))


def movingaverage(arr, window):
    """
    Calculates the moving average ("rolling mean") of an array
    of a certain window size.
    """
    m = np.ones(int(window)) / int(window)
    return scipy.ndimage.convolve1d(arr, m, axis=0, mode='reflect')


def savitzkygolay(arr, window, order, deriv=0):
    # adapted from http://www.scipy.org/Cookbook/SavitzkyGolay
    # but uses ndimage.convolve now, so we don't have to
    # do the padding ourselves
    half_wind = (int(window) - 1) // 2
    order_range = range(int(order) + 1)
    # precompute coefficients
    b = [[k ** i for i in order_range]
         for k in range(-half_wind, half_wind + 1)]
    m = np.linalg.pinv(b)[int(deriv)]
    return scipy.ndimage.convolve1d(arr, m, axis=0, mode='reflect')


def loads(ast_str):
    """
    Create a Trace from a suitably compressed string.
    """
    data = zlib.decompress(ast_str)
    li = struct.unpack('<L', data[0:4])[0]
    lt = struct.unpack('<L', data[4:8])[0]
    n = data[8:8 + li].decode('utf-8')
    t = np.fromstring(data[8 + li:8 + li + lt])
    d = np.fromstring(data[8 + li + lt:])

    return Trace(d, t, name=n)


def dumps(asts):
    """
    Create a compressed string from an Trace.
    """
    d = asts.values.tostring()
    t = asts.index.values.astype(float).tostring()
    lt = struct.pack('<L', len(t))
    i = asts.name.encode('utf-8')
    li = struct.pack('<L', len(i))
    try:  # python 2
        return buffer(zlib.compress(li + lt + i + t + d))
    except NameError:  # python 3
        return zlib.compress(li + lt + i + t + d)


def ts_func(f):
    """
    This wraps a function that would normally only accept an array
    and allows it to operate on a DataFrame. Useful for applying
    numpy functions to DataFrames.
    """
    def wrap_func(df, *args):
        # TODO: should vectorize to apply over all columns?
        return Chromatogram(f(df.values, *args), df.index, df.columns)
    return wrap_func



import struct
import os.path as op
import numpy as np
from aston.trace import Chromatogram
from aston.tracefile import TraceFile


class WatersAutospec(TraceFile):
    """
    Reads *.IDX and *.DAT files from a Waters Autospec.
    Will need to be extended to other Waters files?
    """

    ext = 'IDX'
    traces = ['#ms']

    # def _total_trace(self, twin=None):
    #     fidx = open(self.filename, 'rb')
    #     tme, tic = [], []
    #     while True:
    #         try:
    #             d = struct.unpack('<IHHffhhh', fidx.read(22))
    #             # I: offset in *.DAT file of data chunk
    #             # H: number of data points in data chunk (bytes / 4)
    #             # H:
    #             # f: TIC of data point
    #             # f: time of data point
    #             # h:
    #             # h: 10, 20, 21, 30, 31, 40, or 41?
    #             # h: 80, A0, C0, E0
    #         except struct.error:
    #             break
    #         tme.append(d[4])
    #         tic.append(d[3])
    #     fidx.close()
    #     return TimeSeries(np.array(tic), np.array(tme), \
    #                       ['TIC']).trace(twin=twin)

    @property
    def data(self):
        fidx = open(self.filename, 'rb')
        fdat = open(op.splitext(self.filename)[0] + '.DAT', 'rb')

        ions = set([])
        while True:
            fdat.seek(fdat.tell() + 2)
            try:
                i = struct.unpack('<H', fdat.read(2))[0]
            except struct.error:
                break
            ions.add(i)
        ions = sorted(ions)

        data = []
        tme = []
        while True:
            try:
                idx = struct.unpack('<IHHffhhh', fidx.read(22))
                tme.append(idx[4])
                fdat.seek(idx[0])
                new_line = np.zeros(len(ions))
                d = struct.unpack('<' + int(idx[1] / 4) * 'HH',
                                  fdat.read(idx[1]))
                for i, v in zip(d[1::2], d[0::2]):
                    new_line[ions.index(i)] = v
                data.append(new_line)
            except struct.error:
                break
        fdat.close()
        fidx.close()
        return Chromatogram(np.array(data), np.array(tme), ions)



import binascii
from glob import glob
import inspect
from importlib import import_module
import mimetypes
import os
import os.path as op

from aston.cache import cache


# File types from http://en.wikipedia.org/wiki/Mass_spectrometry_data_format
# and http://www.amdis.net/What_is_AMDIS/AMDIS_Detailed/amdis_detailed.html
# TODO: .FID | Bruker instrument data format
# TODO: .LRP | Shrader/GCMate
# TODO: .MS  | Varian Saturn Files
# TODO: .MS  | HP Benchtop and MS Engines
# TODO: .MS  | Finnigan (GCQ,INCOS and ITDS formats) also *.MI & *.DAT
# TODO: .MSF | Bruker
# TODO: .PKL | MassLynx associated format
# TODO: .RAW | Micromass MassLynx directory format
# TODO: .RAW | PerkinElmer TurboMass file format
# TODO: .SMS | Saturn SMS
# TODO: .WIFF| ABI/Sciex (QSTAR and QTRAP instrument) format
# TODO: .YEP | Bruker instrument data format

mimes = """
image/png	png	8950
application/fasta   fa,fasta,fna,faa                3E
text/csv            csv                             *
application/mzxml	mzxml	*
application/mzml	mzml	*
application/netcdf  cdf                             4344
application/fcs     fcs,lmd                         46435333
application/vnd-sequencing-ab1	ab1,abi	4142
application/vnd-sequencing-scf	scf	2E73
application/vnd-agilent-chemstation-pump	lpmp1.reg	0233
application/vnd-agilent-chemstation-fraction	lafc1fd.reg	0233
application/vnd-agilent-chemstation-flowinject	acqres.reg	0233
application/vnd-agilent-chemstation-lcstat	lcdiag.reg	0233
application/vnd-agilent-chemstation-flowinject	acqres.reg	0233
application/vnd-agilent-masshunter-pump	cappump.cd	*
application/vnd-agilent-masshunter-temp	tcc1.cd	*
application/vnd-agilent-masshunter-acqmethod	acq_method.xml	*
application/vnd-agilent-masshunter-sampleinfo	sample_info.xml	*
application/vnd-agilent-masshunter-msmsscan  bin          0101
application/vnd-agilent-masshunter-dad  sd          *
application/vnd-agilent-chemstation-fid	ch	0238
application/vnd-agilent-chemstation-fid2	ch	0331
application/vnd-agilent-chemstation-ms  ms          0132
application/vnd-agilent-chemstation-mwd  ch          0233
application/vnd-agilent-chemstation-mwd2  ch          0331
application/vnd-agilent-chemstation-dad  uv          0233
application/vnd-agilent-chemstation-dad2  uv          0331
application/vnd-bruker-msms  ami          *
application/vnd-bruker-baf  baf          2400
application/vnd-inficon-hapsite  hps          0403
application/vnd-sciex-wiff	wiff	D0CF
application/vnd-thermo-cf  cf          FFFF
application/vnd-thermo-dxf  dxf          FFFF
application/vnd-thermo-raw  raw          01A1
application/vnd-waters-autospec  idx          *
"""


def get_mimetype(filename, magic_all):
    ft_magic = {}
    ft_ext = {}
    for line in mimes.strip('\n').split('\n'):
        mime, ext, magic = line.split()
        if magic != '*':
            for m in magic.split(','):
                ft_magic[m] = mime

        if ext != '*':
            for e in ext.split(','):
                ft_ext[e] = mime

    # TODO: maybe do some kind of ranking?
    # need to allow multiple filetypes for common magic/extensions
    for i in [4, 2, 1]:
        magic = binascii.b2a_hex(magic_all[:i]).decode('ascii').upper()
        if magic in ft_magic:
            return ft_magic[magic]

    if filename is not None:
        ext = os.path.splitext(filename)[1].lower()[1:]
        if ext in ft_ext:
            return ft_ext[ext]

    return mimetypes.guess_type(filename)[0]


@cache(maxsize=1)
def tfclasses():
    """
    A mapping of mimetypes to every class for reading data files.
    """
    # automatically find any subclasses of TraceFile in the same
    # directory as me
    classes = {}
    mydir = op.dirname(op.abspath(inspect.getfile(get_mimetype)))
    tfcls = {"<class 'aston.tracefile.TraceFile'>",
             "<class 'aston.tracefile.ScanListFile'>"}
    for filename in glob(op.join(mydir, '*.py')):
        name = op.splitext(op.basename(filename))[0]
        module = import_module('aston.tracefile.' + name)
        for clsname in dir(module):
            cls = getattr(module, clsname)
            if hasattr(cls, '__base__'):
                if str(cls.__base__) in tfcls:
                    classes[cls.mime] = cls
    return classes



import numpy as np
from scipy.optimize import leastsq
from aston.trace import Trace
from aston.spectra.isotopes import delta13c_santrock


def ratio_f(pks, r2=45, r1=44):
    # pull out the time and ratio information
    isopks = [pk for pk in pks if pk.info.get('p-type') == 'isostd']
    x = [pk.time() for pk in isopks if pk.area(mz=r1) > 0]
    y = [pk.area(mz=r2) / pk.area(mz=r1)
         for pk in isopks if pk.area(mz=r1) > 0]

    if len(y) < 1:
        return None

    # initial condition for fitting
    p0 = [y[0], 0]

    # function to fit
    def errfunc(p, x, y):
        return p[0] + p[1] * x - y

    # try fitting; if it doesn't work, just use first ratio peak
    try:
        p, succ = leastsq(errfunc, p0, args=(np.array(x), np.array(y)))
    except Exception:
        p = p0

    # construct a function from the fitted data
    def sim_y(t):
        if isinstance(t, (list, tuple, np.ndarray)):
            return np.array(errfunc(p, t, np.zeros(len(t))))
        else:
            return np.array(errfunc(p, t, 0))
    return sim_y


def ratio_series(ts, pks, r2, r1):
    sim_y = ratio_f(pks, r2, r1)
    return Trace(sim_y(ts.index), ts.index,
                 name='{:.1f}/{:.1f}'.format(r2, r1))


def calc_carbon_isotopes(pks, d13cstd, ks='Craig', d18ostd=23.5):
    # get functions to tell us what the ratios of the standards
    # should be over the course of the run
    r45 = ratio_f(pks, r2=45, r1=44)
    r46 = ratio_f(pks, r2=46, r1=44)

    if r45 is None or r46 is None:
        for pk in pks:
            del pk.info['p-d13c']
        return

    for pk in pks:
        # determine the peak areas for our peak and an interpolated standard
        r45sam = pk.area(45) / pk.area(44)
        r46sam = pk.area(46) / pk.area(44)
        r45std = r45(pk.time())
        r46std = r46(pk.time())
        # determine the d13c value and assign it to the peaks info dict
        d13c = delta13c_santrock(r45sam, r46sam, d13cstd,
                                 r45std, r46std, ks, d18ostd)
        pk.info['p-d13c'] = d13c



# -*- coding: utf-8 -*-
import struct
import numpy as np
from aston.trace import Chromatogram
from aston.tracefile import TraceFile, find_offset


class InficonHapsite(TraceFile):
    mime = 'application/vnd-inficon-hapsite'
    traces = ['#ms']

    def _ions(self, f):
        """
        This is a generator that returns the mzs being measured during
        each time segment, one segment at a time.
        """
        outside_pos = f.tell()
        doff = find_offset(f, 4 * b'\xff' + 'HapsSearch'.encode('ascii'))
        # actual end of prev section is 34 bytes before, but assume 1 rec
        f.seek(doff - 62)
        # seek backwards to find the FFFFFFFF header
        while True:
            f.seek(f.tell() - 8)
            if f.read(4) == 4 * b'\xff':
                break
        f.seek(f.tell() + 64)
        nsegments = struct.unpack('<I', f.read(4))[0]
        for _ in range(nsegments):
            # first 32 bytes are segment name, rest are something else?
            f.seek(f.tell() + 96)
            nions = struct.unpack('<I', f.read(4))[0]
            ions = []
            for _ in range(nions):
                # TODO: check that itype is actually a SIM/full scan switch
                i1, i2, _, _, _, _, itype, _ = struct.unpack('<' + 8 * 'I',
                                                             f.read(32))
                if itype == 0:  # SIM
                    ions.append(i1 / 100.)
                else:  # full scan
                    # TODO: this might be a little hacky?
                    #  ideally we would need to know n for this, e.g.:
                    # ions += np.linspace(i1 / 100, i2 / 100, n).tolist()
                    ions += np.arange(i1 / 100., i2 / 100. + 1, 1).tolist()
            # save the file position and load the position
            # that we were at before we started this code
            inside_pos = f.tell()
            f.seek(outside_pos)
            yield ions
            outside_pos = f.tell()
            f.seek(inside_pos)
        f.seek(outside_pos)

    @property
    def data(self):
        # TODO: handle skip mass ranges
        with open(self.filename, 'rb') as f:
            # read in the time segments/mz ranges for the run

            # read in the data itself
            doff = find_offset(f, 4 * b'\xff' + 'HapsScan'.encode('ascii'))
            if doff is None:
                return
            f.seek(doff - 20)
            data_end = doff + struct.unpack('<I', f.read(4))[0] + 55

            f.seek(doff + 56)
            times, abns, mzs = [], [], []
            cur_seg = None
            mz_reader = self._ions(f)
            while f.tell() <= data_end:
                # record info looks like a standard format
                n, t, _, recs, _, seg = struct.unpack('<IiHHHH', f.read(16))
                if cur_seg != seg:
                    # if we've switched segments, update the list of mzs
                    try:
                        cur_mzs = next(mz_reader)
                    except StopIteration:
                        break
                    mzs += set(cur_mzs).difference(mzs)
                    mz_idx = [mzs.index(i) for i in cur_mzs]
                cur_seg = seg

                # just add the new time in
                times.append(t)

                # read the list of abundances
                cur_abns = struct.unpack('<' + 'f' * recs, f.read(4 * recs))
                # convert this list into an array that matches up with
                # whatever mzs we currently have
                empty_row = np.zeros(len(mzs))
                empty_row[mz_idx] = cur_abns
                # add that row into the list
                abns.append(empty_row)

        # convert the time from milliseconds to minutes
        times = np.array(times, dtype=float) / 60000
        # create the data array and populate it
        data = np.zeros((len(times), len(mzs)))
        for i, r in enumerate(abns):
            data[i, 0:len(r)] = r
        return Chromatogram(data, times, mzs)



import numpy as np
from scipy.stats import gaussian_kde
from scipy.optimize import fmin, brentq
from aston.trace.trace import Trace
from aston.trace.math_traces import savitzkygolay


def base(s, slope_tol='5.0', off_tol='5.0'):
    ic = savitzkygolay(s, 5, 3)[0]
    slopes = np.diff(ic) / np.diff(s.index)
    wind = np.array([0.5, 0.5])
    offsets = np.convolve(ic, wind, mode='valid')
    offsets -= slopes * np.convolve(s.index, wind, mode='valid')

    sl_kern = gaussian_kde(slopes)
    base_slope = fmin(lambda x: -sl_kern(x), 0, disp=False)
    k_h = sl_kern(base_slope) * (1.0 - 10 ** -float(slope_tol))
    s1 = brentq(lambda x: sl_kern(x) - k_h, min(slopes), base_slope)
    s2 = brentq(lambda x: sl_kern(x) - k_h, base_slope, max(slopes))

    valid = np.logical_and(slopes > s1, slopes < s2)
    off_kern = gaussian_kde(offsets[valid])
    base_off = fmin(lambda x: -off_kern(x), 0, disp=False)
    k_h = off_kern(base_off) * (1.0 - 10 ** -float(off_tol))
    o1 = brentq(lambda x: off_kern(x) - k_h, min(offsets), base_off)
    o2 = brentq(lambda x: off_kern(x) - k_h, base_off, max(offsets))

    msk = np.logical_and(valid, offsets > o1)
    msk = np.logical_and(valid, offsets < o2)
    new_ic = np.interp(s.index, s.index[msk], ic[msk])
    return Trace(new_ic, s.index)


def base2(ic, t):
    # INSPIRED by Algorithm A12 from Zupan
    # 5 point pre-smoothing
    sc = np.convolve(np.ones(5) / 5.0, ic, mode='same')
    # get local minima
    mn = np.arange(len(ic))[np.r_[True, ((sc < np.roll(sc, 1)) & (sc < np.roll(sc, -1)))[1:-1], True]]  # noqa
    # don't allow baseline to have a slope greater than
    # 10x less than the steepest peak
    max_slope = np.max(np.gradient(ic)) / 10.0
    slope = max_slope
    pi = 0  # previous index
    oc = np.zeros(len(ic))
    for i in range(1, len(mn)):
        if slope < ((ic[mn[i]] - ic[mn[pi]]) /
                    (mn[i] - mn[pi])) and slope < max_slope:
            # add trend
            oc[mn[pi]:mn[i - 1]] = np.linspace(ic[mn[pi]], ic[mn[i - 1]],
                                               mn[i - 1] - mn[pi])
            pi = i - 1
        slope = (ic[mn[i]] - ic[mn[pi]]) / (mn[i] - mn[pi])
    print(mn[pi], mn[-1])
    oc[mn[pi]:mn[-1]] = np.linspace(ic[mn[pi]], ic[mn[-1]], mn[-1] - mn[pi])
    oc[-1] = oc[-2]  # FIXME: there's definitely a bug in here somewhere



import numpy as np


class Scan(object):
    def __init__(self, x, abn, name='', source=None):
        assert len(x) == len(abn)
        self.x, self.abn = x, abn
        self.name = name
        self.source = source

    def plot(self, color=None, label=False, ax=None):
        # TODO: this is extremely ugly; needs a rewrite
        # TODO: should use source?
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if color is None:
            color = 'k'

        scn = np.vstack([self.x, self.abn])

        # determine if the spectrum is continuous
        cont_spec = scn.shape[1] > 10
        cont_spec = cont_spec and np.all(np.abs(np.diff(scn[0]) -
                                                (scn[0, 1] - scn[0, 0])) < 1e-9)  # noqa

        if cont_spec:
            # if the spacing between all the points is equal, plot as a line
            scn = scn[:, np.argsort(scn)[0]]
            ax.plot(scn[0], scn[1], '-', color=color)
        else:
            # remove 0's
            scn = scn[:, scn[1] != 0]
            try:
                # FIXME: this crashes on Windows unless the user has
                # clicked on the spectrum graph previously.
                # Matplotlib bug needs workaround
                ax.vlines(scn[0], 0, scn[1], color=color, alpha=0.5)
            except Exception:
                pass
            ax.plot(scn[0], scn[1], ',', color=color)

        cut_val = 0.01 * max(np.array(scn[1]))  # only label peaks 1% of max
        if label and cont_spec:
            l_idxs = []
            for idx in scn[1].argsort()[::-1]:
                l_idxs.append(idx)
                if idx - 1 in l_idxs or idx + 1 in l_idxs:
                    continue
                s = scn[:, idx].T
                # TODO: allow a definable max number of labels?
                # or do better collision detection between them?
                if s[1] < cut_val or len(l_idxs) > 100:
                    break
                ax.text(s[0], s[1], str(s[0]), ha='center', va='bottom',
                        rotation=90, size=10, color=color)
        elif label and not cont_spec:
            filt_scn = scn[:, 0.5 * np.roll(scn[1], 1) - scn[1] <= 0]
            for s in filt_scn[:, filt_scn[1] > cut_val].T:
                ax.text(s[0], s[1], str(s[0]), ha='center', va='bottom',
                        rotation=90, size=10, color=color)

        #    #go through the top 10% highest ions from highest to lowest
        #    #always have at least 10 labels, but no more than 50 (arbitrary)
        #    #if an ion is close to one seen previously, don't display it
        #    v2lbl = {}  # values to label
        #    plbl = []  # all values so far
        #    max_val = max(np.array(scn[1]))  # only label peaks X % of this
        #    for i in np.array(scn[1]).argsort()[::-1]:
        #        mz = scn[0][i]
        #        #don't allow a new label within 1.5 units of another
        #        if not np.any(np.abs(np.array(plbl) - mz) < 1.5) and \
        #          scn[1][i] > 0.01 * max_val:
        #            v2lbl[mz] = scn[1][i]
        #        plbl.append(mz)

        #    #add peak labels
        #    for v in v2lbl:
        #        ax.text(v, v2lbl[v], str(v), ha='center', \
        #            va='bottom', rotation=90, size=10, color=clr)
        #        #ax.text(v, v2lbl[v], str(v), ha='center', \
        #        #    va='bottom', rotation=90, size=10, color=clr, \
        #        #    bbox={'boxstyle': 'larrow,pad=0.3', 'fc': clr, \
        #        #        'ec': clr, 'lw': 1, 'alpha': '0.25'})

    @property
    def xmin(self):
        return min(self.x)

    @property
    def xmax(self):
        return max(self.x)

    @property
    def ymin(self):
        return min(self.abn)

    @property
    def ymax(self):
        return max(self.abn)

    def d13c(self):
        # FIXME: this needs to be moved to somewhere else;
        # can't get parent in here
        if self.source != 'irms':
            return None
        pass

        # dt = self.getParentOfType('file')
        # if self.info['sp-type'] == 'Isotope Standard':
        #     return dt.info['r-d13c-std']

        # # if there's no reference number, we can't do this
        # try:
        #     float(dt.info['r-d13c-std'])
        # except:
        #     return ''

        # r45std = dt.get_point('r45std', float(self.info['sp-time']))
        # r46std = dt.get_point('r46std', float(self.info['sp-time']))

        # # if no peak has been designated as a isotope std
        # if r45std == 0.0:
        #     return ''

        # d = delta13c_santrock(self.ion(44), self.ion(45), self.ion(46), \
        #          float(dt.info['r-d13c-std']), r45std, r46std)

        # return str(d)



import struct
import re
import os
import numpy as np
from aston.trace import Chromatogram
from aston.tracefile import TraceFile, find_offset


class ThermoCF(TraceFile):
    mime = 'application/vnd-thermo-cf'
    traces = ['#irms']

    @property
    def data(self):
        f = open(self.filename, 'rb')
        f.seek(19)
        while True:
            f.seek(f.tell() - 19)
            if f.read(19) == b'CRawDataScanStorage':
                break
            if f.read(1) == b'':
                f.close()
                return

        f.seek(f.tell() + 62)
        nscans = struct.unpack('H', f.read(2))[0]

        # TODO: this shouldn't be hardcoded
        ions = [44, 45, 46]
        ni = len(ions)

        f.seek(f.tell() + 35)
        data = np.array([struct.unpack('<f' + ni * 'd', f.read(4 + ni * 8))
                         for _ in range(nscans)])
        data[:, 0] /= 60.  # convert time to minutes
        # self.data = TimeSeries(data[:, 1:], data[:, 0], ions)
        f.close()
        return Chromatogram(data[:, 1:], data[:, 0], ions)

    @property
    def info(self):
        d = super(ThermoCF, self).info
        # info['file name'] = os.path.basename(self.filename)
        d['name'] = os.path.splitext(os.path.basename(self.filename))[0]
        return d


class ThermoDXF(TraceFile):
    mime = 'application/vnd-thermo-dxf'
    traces = ['#irms', '*refgas']

    @property
    def data(self):
        f = open(self.filename, 'rb')

        # TODO: use find_offset to find this?
        # f.seek(11)
        # while True:
        #     f.seek(f.tell() - 11)
        #     if f.read(11) == b'CEvalGCData':
        #         break
        #     if f.read(1) == b'':
        #         f.close()
        #         return
        f.seek(find_offset(f, b'CRawData') + 9)
        strlen = 2 * struct.unpack('<B', f.read(1))[0]
        tname = f.read(strlen).decode('utf_16_le')
        if tname == 'CO2':
            ions = [44, 45, 46]
        elif tname == 'CO':
            ions = [28, 29, 30]
        elif tname == 'SO2,SO-SO2 Ext,SO':
            # TODO: check this is in the right order
            ions = [48, 49, 50, 64, 65, 66]
        else:
            # TODO: should save the tname somewhere for future reference
            ions = [1, 2, 3]

        f.seek(find_offset(f, b'CEvalGCData') + 4)

        # bytes until the end converted to # of records
        nscans = int(struct.unpack('<I', f.read(4))[0] /
                     (4.0 + len(ions) * 8.0))

        dtype = np.dtype([('index', '<f4'), ('values', '<f8', len(ions))])
        data = np.fromfile(f, dtype=dtype, count=nscans)
        # convert time to minutes
        data['index'] /= 60.

        f.close()
        return Chromatogram(data['values'], data['index'], ions)

    @property
    def info(self):
        d = super(ThermoDXF, self).info
        # try: #TODO: this crashes in python 3; not clear why?
        # except:
        #     pass
        # info['file name'] = os.path.basename(self.filename)
        d['name'] = os.path.splitext(os.path.basename(self.filename))[0]
        with open(self.filename, 'rb') as f:
            foff_o = find_offset(f, 'd 18O/16O'.encode('utf_16_le'))
            foff_c = find_offset(f, 'd 13C/12C'.encode('utf_16_le'))
            if foff_o is not None:
                f.seek(foff_o + 68)
                d['d18o_std'] = str(struct.unpack('<d', f.read(8))[0])
            if foff_c is not None:
                f.seek(foff_c + 68)
                d['d13c_std'] = str(struct.unpack('<d', f.read(8))[0])
        return d

    def _th_off(self, search_str, hint=None):
        # TODO: replace all occurances with find_offset?
        if hint is None:
            hint = 0
        with open(self.filename, 'rb') as f:
            f.seek(hint)
            regexp = re.compile(search_str.encode('ascii'))
            while True:
                d = f.read(len(search_str) * 200)
                srch = regexp.search(d)
                if srch is not None:
                    foff = f.tell() - len(d) + srch.end()
                    break
                if len(d) == len(search_str):  # no data read: EOF
                    f.close()
                    return None
                f.seek(f.tell() - len(search_str))
        return foff

    def events(self, name='refgas', twin=None):
        # TODO: use twin
        # TODO: read in ref gas pulses
        evts = super(ThermoDXF, self).events(name, twin)
        if name != 'refgas':
            return evts

        with open(self.filename, 'rb') as f:
            f.seek(self._th_off('CActionHwTransferContainer'))
            evt, time, status = [], [], []
            while True:
                d = struct.unpack('<ihHBB', f.read(10))
                if d[0] != 3:
                    # TODO: probably a better way to figure this out
                    break
                # name1 = f.read(2 * d[4]).decode('utf-16')
                f.read(2 * d[4])
                d = struct.unpack('<HBB', f.read(4))
                name2 = f.read(2 * d[2]).decode('utf-16')
                d = struct.unpack('<iiiiiiiih', f.read(34))
                if d[4] == 2:
                    status.append(d[5])
                else:
                    # TODO: needs a correction factor to be
                    # derived from guesswork; timing in
                    # Isodat/Conflo is uncoupled? BAD
                    time.append(d[4] / 60000.)
                    evt.append(name2)
                if d[-1] == -1:
                    # first record ends (?) with this, so skip
                    # the CValveTransfer part afterwards
                    f.read(22)
        p_st, gas_on, i = None, False, 0
        for e, t, s in zip(evt, time, status):
            if e.startswith('Reference'):
                gas_on = (s == 1)
                if gas_on:
                    p_st = t
                elif not gas_on and p_st is not None:
                    i += 1
                    evts.append({'t0': p_st, 't1': t, 'name': 'R' + str(i)})
        return evts


class ThermoRAW(TraceFile):
    mime = 'application/vnd-thermo-raw'
    traces = ['#ms']

    # TODO: write the rest of this!



import numpy as np
from aston.peak.peak_finding import simple_peak_find
from aston.peak.integrators import simple_integrate


def _get_windows(peak_list):
    """
    Given a list of peaks, bin them into windows.
    """
    win_list = []
    for t0, t1, hints in peak_list:
        p_w = (t0, t1)
        for w in win_list:
            if p_w[0] <= w[0][1] and p_w[1] >= w[0][0]:
                w[0] = (min(p_w[0], w[0][0]), max(p_w[1], w[0][1]))
                w[1].append((t0, t1, hints))
                break
        else:
            win_list.append([p_w, [(t0, t1, hints)]])
    return win_list


def integrate(s, peak_list=None, baseline=None, overlap='drop'):
    """

    """
    # for convenience, if no peak_list is provided create our own
    if peak_list is None:
        peak_list = simple_peak_find(s)

    peaks = []
    for _, pks in _get_windows(peak_list):
        temp_pks = []
        pks = sorted(pks, key=lambda p: p[0])
        if 'y0' in pks[0][2] and 'y1' in pks[-1][2]:
            y0, y1 = pks[0][2]['y0'], pks[-1][2]['y1']
        else:
            y0 = s.get_point('!', pks[0][0])
            y1 = s.get_point('!', pks[-1][1])
        ys = np.array([y0, y1])
        xs = np.array([pks[0][0], pks[-1][1]])

        # go through list of peaks to make sure there's no overlap
        for t0, t1, hints in pks:
            # figure out the y values (using a linear baseline)
            hints['y0'] = np.interp(t0, xs, ys)
            hints['y1'] = np.interp(t1, xs, ys)

            # if this peak totally overlaps with an existing one, don't add
            if sum(1 for p in temp_pks if t1 <= p[1]) > 0:
                continue
            overlap_pks = [p for p in temp_pks if t0 <= p[1]]
            if len(overlap_pks) > 0:
                # find the last of the overlapping peaks
                overlap_pk = max(overlap_pks, key=lambda p: p[0])
                # get the section of trace and find the lowest point
                over_ts = s.trace('!', twin=(t0, overlap_pk[1]))
                min_t = over_ts.times[over_ts.y.argmin()]

                # delete the existing overlaping peak
                for i, p in enumerate(temp_pks):
                    if p == overlap_pk:
                        del temp_pks[i]
                        break

                # interpolate a new y value
                y_val = np.interp(min_t, xs, ys)
                overlap_pk[2]['y1'] = y_val
                hints['y0'] = y_val

                # add the old and new peak in
                temp_pks.append((overlap_pk[0], min_t, overlap_pk[2]))
                temp_pks.append((min_t, t1, hints))
            else:
                temp_pks.append((t0, t1, hints))

        # none of our peaks should overlap, so we can just use
        # simple_integrate now
        peaks += simple_integrate(s, temp_pks)
    for p in peaks:
        p.info['p-create'] = p.info['p-create'].split(',')[0] + \
            ',drop_integrate'
    return peaks


def integrate_analytical(s, foundpeaks=None, baseline=None):
    pass



import numpy as np
from aston.spectra import Scan


def read_amdis(filename):
    def make_spc(arr_str, info):
        b = [int(i.strip('()')) for i in
             arr_str.split() if i.strip('()') != '']
        b = np.array(b)
        data = np.reshape(b, (b.shape[0] / 2, 2)).T
        return Scan(info, data)

    def val_prt(x):
        return x.split(':')[1].lower().strip()

    with open(filename, 'r') as f:
        info, arr_str = {'name': val_prt(f.readline())}, ''
        for line in f:
            if line.startswith('('):
                arr_str += line + ' '
            elif line.startswith('NAME:'):
                spc = make_spc(arr_str, info)
                yield spc
                info, arr_str = {'name': val_prt(line)}, ''
            elif line.startswith('RT:'):
                info['p-s-time'] = val_prt(line)


def read_jcamp(filename):
    raise NotImplementedError



#!/usr/bin/env python

from setuptools import setup
import os

from aston import __version__


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

# read in the version number
with open('aston/__init__.py') as f:
    exec(f.read())

options = {
    'name': 'Aston',
    'version': __version__,
    'description': 'Mass/UV Spectral Analysis Program',
    'author': 'Roderick Bovee',
    'author_email': 'rbovee@gmail.com',
    'url': 'https://github.com/bovee/aston',
    'license': 'BSD 3-Clause',
    'platforms': ['Any'],
    'classifiers': [
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Topic :: Scientific/Engineering :: Chemistry'
    ],
    'long_description': read('README.md'),
    'long_description_content_type': 'text/markdown',
    'packages': [
        'aston', 'aston.calibrations', 'aston.peak',
        'aston.spectra', 'aston.trace', 'aston.tracefile'
    ],
    'scripts': [],
    # 'data_files': matplotlib.get_py2exe_datafiles(),
    'package_data': {'aston': []},
    'include_package_data': True,
    'install_requires': ['numpy>=1.16.4', 'scipy>=1.2.0'],
    'setup_requires': ['pytest-runner'],
    'tests_require': ['pytest', 'coverage', 'pytest-cov'],
    'extras_require': {
        'plot': ['matplotlib', 'jupyter'],
    }
}

# all the magic happens right here
setup(**options)



'''
Classes that can open chromatographic files and return
info from them or Traces/Chromatograms.
'''

import re
import struct

import numpy as np

from aston.trace import Chromatogram, Trace
from aston.tracefile.mime import get_mimetype, tfclasses


def find_offset(f, search_str, hint=None):
    if hint is None:
        hint = 0
    f.seek(hint)
    regexp = re.compile(search_str)
    while True:
        d = f.read(len(search_str) * 200)
        srch = regexp.search(d)
        if srch is not None:
            foff = f.tell() - len(d) + srch.end()
            break
        if len(d) == len(search_str):  # no data read: EOF
            return None
        f.seek(f.tell() - len(search_str))
    return foff


def parse_c_serialized(f):
    """
    Reads in a binary file created by a C++ serializer (prob. MFC?)
    and returns tuples of (header name, data following the header).
    These are used by Thermo for *.CF and *.DXF files and by Agilent
    for new-style *.REG files.
    """
    # TODO: rewrite to use re library
    f.seek(0)
    try:
        p_rec_type = None
        while True:
            rec_off = f.tell()
            while True:
                if f.read(2) == b'\xff\xff':
                    h = struct.unpack('<HH', f.read(4))
                    if h[1] < 64 and h[1] != 0:
                        rec_type = f.read(h[1])
                        if rec_type[0] == 67:  # starts with 'C'
                            break
                if f.read(1) == b'':
                    raise EOFError
                f.seek(f.tell() - 2)
            if p_rec_type is not None:
                rec_len = f.tell() - 6 - len(rec_type) - rec_off
                f.seek(rec_off)
                yield p_rec_type, f.read(rec_len)
                f.seek(f.tell() + 6 + len(rec_type))
            # p_type = h[0]
            p_rec_type = rec_type
    except EOFError:
        rec_len = f.tell() - 6 - len(rec_type) - rec_off
        f.seek(rec_off)
        yield p_rec_type, f.read(rec_len)


class TraceFile(object):
    mime = ''  # mimetype to associate file with (in tracefile.mime)

    # traces is a list of possible traces:
    # each item may start with a * indicating a events
    # a # indicating a 2d trace or nothing indicating a single trace name
    traces = []

    def __init__(self, filename=None, ftype=None, data=None):
        self.filename = filename
        self.ftype = ''
        self._data = None

        # try to automatically change my class to reflect
        # whatever type of file I'm pointing at if not provided
        if type(self) is TraceFile:
            if data is not None:
                self._data = data

            if filename is None:
                return

            with open(filename, mode='rb') as f:
                magic = f.read(4)

            ftype = get_mimetype(filename, magic)

            if ftype is not None:
                if ftype in tfclasses():
                    self.__class__ = tfclasses()[ftype]
                    self.ftype = ftype
        else:
            self.ftype = self.__class__.__name__

    @property
    def data(self):
        if self._data is not None:
            return self._data
        else:
            return Chromatogram()

    def scans(self):
        # TODO: decompose self.data into scans
        pass

    def total_trace(self, twin=None):
        return self.data.trace(twin=twin)

    # TODO: should this code be kept? (needs to be improved, if so)
    # def plot(self, name='', ax=None):
    #     if ax is None:
    #         import matplotlib.pyplot as plt
    #         ax = plt.gca()

    #     for t in self.traces:
    #         if t.startswith('#'):
    #             #self.trace(t[1:]).plot(ax=ax)
    #             self.trace('').plot(ax=ax)
    #         elif t.startswith('*'):
    #             #TODO: plot events
    #             pass
    #         else:
    #             self.trace(t).plot(ax=ax)

    #     #TODO: colors?
    #     #TODO: plot 2d/colors

    def trace(self, name='', tol=0.5, twin=None):
        if isinstance(name, (int, float, np.float32, np.float64)):
            name = str(name)
        else:
            name = name.lower()

        # name of the 2d trace, if it exists
        if any(t.startswith('#') for t in self.traces):
            t2d = [t[1:] for t in self.traces if t.startswith('#')][0]

            # clip out the starting 'MS' if present
            if name.startswith(t2d):
                name = name[len(t2d):]
        else:
            t2d = ''

        # this is the only string we handle; all others handled in subclasses
        if name in ['tic', 'x', '']:
            return self.total_trace(twin)
        elif name in self.traces:
            return self._trace(name, twin)
        elif t2d != '':
            # this file contains 2d data; find the trace in that
            return self.data.trace(name, tol, twin)
        else:
            return Trace()

    def scan(self, t, dt=None, aggfunc=None):
        """
        Returns the spectrum from a specific time or range of times.
        """
        return self.data.scan(t, dt, aggfunc)

    @property
    def info(self):
        # TODO: add creation date and short name
        return {'filename': self.filename,
                'filetype': self.ftype}

    def events(self, name, twin=None):
        # TODO: check for '*' trace in self.traces
        return []

    def subscan(self, name, t, mz):
        """
        Returns a spectra linked to both a time and mz, e.g.
        the daughter scan in an MSMS or a MS scan from a GC-GC.
        """
        pass

    def subscans(self, name, twin=None):
        """
        Returns a list of times with subscans and their associated mzs.

        Preliminary idea:
        If all points in self.data have subscans, return True.
        """
        # example: [0.1], [147, 178]
        return [], []

    def md5hash(self):
        # TODO: calculate md5hash of this file
        # to be used for determining if files in db are unique
        raise NotImplementedError


class ScanListFile(TraceFile):
    def scans(self, twin=None):
        return []

    # TODO: is there a point in creating a data property here? (for heatmaps?)
    # TODO: if so, then need better binning code...

    def total_trace(self, twin=None):
        if twin is None:
            twin = (-np.inf, np.inf)
        times, y = [], []
        for s in self.scans(twin):
            t = float(s.name)
            if t < twin[0]:
                continue
            if t > twin[1]:
                break
            times.append(t)
            y.append(sum(s.abn))
        return Trace(y, times, name='tic')

    def trace(self, name='', tol=0.5, twin=None):
        if twin is None:
            twin = (-np.inf, np.inf)
        if name in {'tic', 'x', ''}:
            return self.total_trace(twin)
        times, y = [], []
        for s in self.scans(twin):
            t = float(s.name)
            if t < twin[0]:
                continue
            if t > twin[1]:
                break
            times.append(t)
            # TODO: this can be vectorized with numpy?
            y.append(sum(j for i, j in zip(s.x, s.abn)
                         if np.abs(i - name) < tol))
        return Trace(y, times, name=name)

    def scan(self, t, dt=None, aggfunc=None):
        # TODO: use aggfunc
        prev_s = None
        bin_scans = []
        for s in self.scans():
            if float(s.name) > t:
                if float(prev_s.name) - t < float(s.name) - t:
                    if dt is None:
                        return prev_s
                    else:
                        bin_scans.append(prev_s)
                elif dt is None:
                    return s
                bin_scans.append(s)
                if float(s.name) > t + dt:
                    break
            prev_s = s
        # merge bin_scans and return
        # FIXME
        pass
